#include "collidable.hpp"

#include <iostream>
#include <numbers>
#include <numeric>

// /**
//  * @brief Intersection test between two segments [AB] and [CD]
//  *
//  * Inspired from https://nlecam.pages.unistra.fr/geo3d_intersect/
//  *
//  * @param a Point A
//  * @param b Point B
//  * @param c Point C
//  * @param d Point D
//  * @return A boolean indicating whether the segments intersects
//  */
// bool is_intersecting(Vector2D const &a, Vector2D const &b, Vector2D const &c, Vector2D const &d)
// {
//     auto determinant = [](Vector2D const &p1, Vector2D const &p2, Vector2D const &p3)
//     {
//         return (p2.x - p1.x) * (p3.y - p1.y) - (p2.y - p1.y) * (p3.x - p1.x);
//     };

//     float s1 = determinant(a, b, c);
//     float s2 = determinant(a, b, d);
//     float s3 = determinant(d, c, a);
//     float s4 = determinant(d, c, b);

//     auto between = [](float x, float x0, float x1)
//     {
//         return x >= x0 && x <= x1;
//     };

//     if (s1 * s2 == 0 && s3 * s4 == 0)
//     {
//         Vector2D ab = b - a;
//         Vector2D ac = c - a;
//         Vector2D ad = d - a;

//         Vector2D cd = d - c;
//         Vector2D ca = a - c;
//         Vector2D cb = b - c;

//         return (
//             between(ac.dot(ab) / ab.dot(ab), 0.0, 1.0) ||
//             between(ad.dot(ab) / ab.dot(ab), 0.0, 1.0) ||
//             between(ca.dot(cd) / cd.dot(cd), 0.0, 1.0) ||
//             between(cb.dot(cd) / cd.dot(cd), 0.0, 1.0));
//     }

//     return s1 * s2 <= 0 && s3 * s4 <= 0;
// }

void CircleShape::debug_draw(SDL2::Renderer::SP canvas) const
{
    int const LOD = 16;
    float const INVERSE_LOD = 1.0f / LOD;

    float angle = 0.0f;
    Vector2D previous = radius * Vector2D{std::cos(angle), std::sin(angle)} + origin;

    for (size_t i = 1; i <= LOD; ++i)
    {
        angle = i * INVERSE_LOD * 2.0f * std::numbers::pi;
        Vector2D current = radius * Vector2D{std::cos(angle), std::sin(angle)} + origin;
        SDL_SetRenderDrawColor(canvas.get(), 255, 100, 0, 255);
        SDL_RenderDrawLine(canvas.get(), previous.x, previous.y, current.x, current.y);
        previous = current;
    }
}

bool CircleShape::is_overlaping(PolygonShape const &poly) const
{
    float const squared_radius = radius * radius;
    bool is_all_inside = true;

    Vector2D previous = poly.vertices().front();

    for (Vector2D &&v : poly.vertices() | std::views::drop(1))
    {
        if (origin.squared_distance_to_segment(previous, v) <= squared_radius)
            return true;

        Vector2D midpoint = (previous + v) / 2.0f;
        is_all_inside &= (v - previous).orthogonal().dot(origin - midpoint) < 0.0;

        previous = v;
    }

    Vector2D const &first = poly.vertices().front();
    Vector2D const &last = poly.vertices().back();

    if (origin.squared_distance_to_segment(last, first) <= squared_radius)
        return true;

    Vector2D midpoint = (last + first) / 2.0f;
    is_all_inside &= (first - last).orthogonal().dot(origin - midpoint) < 0.0;

    return is_all_inside;
}

bool CircleShape::is_overlaping(CircleShape const &circle) const
{
    float radius_sum = (radius + circle.radius);
    return (origin - circle.origin).squared_length() <= radius_sum * radius_sum;
}

Vector2D CircleShape::get_overlap_normal(PolygonShape const &poly) const
{
    float const squared_radius = radius * radius;

    Vector2D normal_sum;
    Vector2D previous = poly.vertices().front();

    for (Vector2D &&v : poly.vertices() | std::views::drop(1))
    {
        if (origin.squared_distance_to_segment(previous, v) <= squared_radius)
            normal_sum += (v - previous).orthogonal().normalized();

        previous = v;
    }

    Vector2D const &first = poly.vertices().front();
    Vector2D const &last = poly.vertices().back();

    if (origin.squared_distance_to_segment(last, first) <= squared_radius)
        normal_sum += (first - last).orthogonal().normalized();

    return normal_sum.normalized();
}

Vector2D CircleShape::get_overlap_normal(CircleShape const &circle) const
{
    return (origin - circle.origin).normalized();
}

/*
  PolygonShape
*/

void PolygonShape::debug_draw(SDL2::Renderer::SP canvas) const
{
    if (vertices_.size() < 2)
    {
        return;
    }

    Vector2D first = transform * vertices_[0];
    Vector2D previous = first;
    for (size_t i = 1; i < vertices_.size(); i++)
    {
        Vector2D current = transform * vertices_[i];
        SDL_SetRenderDrawColor(canvas.get(), 255, 0, 0, 255);
        SDL_RenderDrawLine(canvas.get(), previous.x, previous.y, current.x, current.y);
        previous = current;
    }

    SDL_RenderDrawLine(canvas.get(), previous.x, previous.y, first.x, first.y);
}

bool PolygonShape::is_overlaping(CircleShape const &circle) const
{
    return circle.is_overlaping(*this);
}

bool PolygonShape::is_overlaping(PolygonShape const &circle) const
{
    throw std::logic_error("Not implemented");
}

Vector2D PolygonShape::get_overlap_normal(CircleShape const &circle) const
{
    throw std::logic_error("Not implemented");
}

Vector2D PolygonShape::get_overlap_normal(PolygonShape const &circle) const
{
    throw std::logic_error("Not implemented");
}



AABB PolygonShape::get_aabb() const
{
    Vector2D top_left = vertices().front();
    Vector2D bottom_right = vertices().front();

    for (auto &&v : vertices())
    {
        top_left = {std::min(v.x, top_left.x), std::min(v.y, top_left.y)},
        bottom_right = {std::max(v.x, bottom_right.x), std::max(v.y, bottom_right.y)};
    }

    return {top_left, bottom_right};
}

std::unique_ptr<PolygonShape> PolygonShape::create_regular(int n_vertices)
{
    float const INVERSE_N = 1.0f / n_vertices;

    std::vector<Vector2D> points(n_vertices);

    float angle = 0.0f;
    for (size_t i = 0; i < n_vertices; ++i)
    {
        angle = i * INVERSE_N * 2.0f * std::numbers::pi;
        points[i] = {std::cos(angle), std::sin(angle)};
    }

    return std::make_unique<PolygonShape>(std::move(points));
}

void AABB::debug_draw(SDL2::Renderer::SP canvas) const
{
    SDL_SetRenderDrawColor(canvas.get(), 0, 0, 255, 255);
    SDL_RenderDrawLine(canvas.get(), top_left.x, top_left.y, top_left.x, bottom_right.y);
    SDL_RenderDrawLine(canvas.get(), top_left.x, top_left.y, bottom_right.x, top_left.y);
    SDL_RenderDrawLine(canvas.get(), bottom_right.x, top_left.y, bottom_right.x, bottom_right.y);
    SDL_RenderDrawLine(canvas.get(), top_left.x, bottom_right.y, bottom_right.x, bottom_right.y);
}

bool Collidable::collide_movement(Collidable const &other, Vector2D &movement, Vector2D &collision_normal)
{
    if(!aabb_.is_intersecting(other.aabb_))
        return false;

    bool collided = false;
    float const delta = 0.05;

    collision_shape_->move(movement);

    int resolve_steps = 0;
    while (is_colliding_with(other))
    {
        collided = true;
        collision_shape_->move(movement * -delta);
        resolve_steps++;
    }
    collision_shape_->move(movement * delta);

    if (auto circle = dynamic_cast<CircleShape *>(other.collision_shape_.get()))
    {
        collision_normal = collision_shape_->get_overlap_normal(*circle);
    }
    else if (auto polygon = dynamic_cast<PolygonShape *>(other.collision_shape_.get()))
    {
        collision_normal = collision_shape_->get_overlap_normal(*polygon);
    }

    collision_shape_->move(movement * -delta);


    movement -= movement * resolve_steps * delta;
    collision_shape_->move((movement * resolve_steps * (1 - delta)));

    return collided;
}