#pragma once

#include <vector>
#include <memory>
#include <iostream>
#include <ranges>

#include "interfaces.hpp"
#include "vector2d.hpp"
#include "transform2d.hpp"
#include "sdl.hpp"

class PolygonShape;
class CircleShape;

class AABB : public DebugDraw
{
public:
    Vector2D top_left;
    Vector2D bottom_right;

    AABB(Vector2D top_left, Vector2D bottom_right) : top_left(top_left), bottom_right(bottom_right) {}

    void debug_draw(SDL2::Renderer::SP renderer) const override;

    inline bool contains(Vector2D const &v)
    {
        return top_left.x <= v.x && bottom_right.x >= v.x && top_left.y <= v.y && bottom_right.y >= v.y;
    }

    inline bool is_intersecting(AABB const &other)
    {
        return !(top_left.y > other.bottom_right.y || bottom_right.y < other.top_left.y || top_left.x > other.bottom_right.x || bottom_right.x < other.top_left.x);
    }
};

class CollisionShape : public DebugDraw
{
public:
    virtual bool is_overlaping(CollisionShape const &shape) const = 0;
    virtual bool is_overlaping(PolygonShape const &poly) const = 0;
    virtual bool is_overlaping(CircleShape const &shape) const = 0;
    virtual AABB get_aabb() const = 0;
    virtual void move(Vector2D const &v) = 0;
    virtual Vector2D get_overlap_normal(PolygonShape const &poly) const = 0;
    virtual Vector2D get_overlap_normal(CircleShape const &circle) const = 0;
};

template <typename T>
class OverlapMixin : public CollisionShape
{
    bool is_overlaping(CollisionShape const &other) const override
    {
        return other.is_overlaping(static_cast<T const &>(*this));
    }
};

class CircleShape : public OverlapMixin<CircleShape>
{
public:
    float radius;
    Vector2D origin;

    CircleShape(Vector2D origin, float radius) : origin(origin), radius(radius) {}
    void debug_draw(SDL2::Renderer::SP renderer) const override;
    bool is_overlaping(PolygonShape const &poly) const override;
    bool is_overlaping(CircleShape const &circle) const override;

    Vector2D get_overlap_normal(PolygonShape const &poly) const override;
    Vector2D get_overlap_normal(CircleShape const &circle) const override;

    inline AABB get_aabb() const override
    {
        return {origin + radius * Vector2D{-1, -1}, origin + radius * Vector2D{1, 1}};
    }

    inline void move(Vector2D const &v) override
    {
        origin += v;
    }
};

class PolygonShape : public OverlapMixin<PolygonShape>
{
private:
    std::vector<Vector2D> vertices_;

public:
    Transform2D transform;

    PolygonShape(std::initializer_list<Vector2D> vertices) : vertices_(vertices) {}
    PolygonShape(std::vector<Vector2D> &&vertices) : vertices_(vertices) {}

    inline auto vertices() const
    {
        return (vertices_ |
                std::views::transform([&](auto const v)
                                      { return transform * v; }));
    }

    void debug_draw(SDL2::Renderer::SP canvas) const override;
    bool is_overlaping(CircleShape const &circle) const override;
    bool is_overlaping(PolygonShape const &circle) const override;

    Vector2D get_overlap_normal(PolygonShape const &poly) const override;
    Vector2D get_overlap_normal(CircleShape const &circle) const override;

    static std::unique_ptr<PolygonShape> create_regular(int n_vertices);
    
    AABB get_aabb() const override;
    
    inline void move(Vector2D const &v) override
    {
        transform = Transform2D::translation(v) * transform;
    }
};


class Collidable
{
protected:
    AABB aabb_;
    std::unique_ptr<CollisionShape> collision_shape_;

public:
    Collidable(std::unique_ptr<CollisionShape> collision_shape) : collision_shape_(std::move(collision_shape)), aabb_{collision_shape->get_aabb()} {}

    inline bool is_colliding_with(Collidable const &other)
    {
        return aabb_.is_intersecting(other.aabb_) && collision_shape_->is_overlaping(*other.collision_shape_);
    }

    /**
     * @brief Performs a move with a collision check with @p other
     * 
     * @param other The collidable to test the collision with
     * @param[in, out] movement The desired move to make, modified to avoid the collision if any
     * @param[out] collision_normal The normal of the collision if any
     * @return true if there was a collision
     */
    bool collide_movement(Collidable const &other, Vector2D &movement, Vector2D &collision_normal);
};