#pragma once

#include <SDL2/SDL.h>
#include <SDL2/SDL_image.h>

#include <filesystem>
#include <memory>
#include <format>

#include "input.hpp"

/*
 * smart pointer wrappers credit: https://swarminglogic.com/scribble/2015_05_smartwrappers
 */

namespace SDL2
{
    struct Guard
    {
        Guard()
        {
            if (SDL_Init(SDL_INIT_VIDEO) != 0)
            {
                throw std::runtime_error(std::format("Failed to initialize SDL: {}", SDL_GetError()));
            }

            if (!IMG_Init(IMG_INIT_PNG))
            {
                throw std::runtime_error(std::format("Failed to initialize SDL_image: {}", SDL_GetError()));
            }

            // Pretty anti-aliasing
            SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "2");

            Input::initialize();
        }

        ~Guard()
        {
            SDL_Quit();
            IMG_Quit();
        }
    };

    struct Deleter
    {
        void operator()(SDL_Surface *ptr)
        {
            if (ptr)
                SDL_FreeSurface(ptr);
        }
        void operator()(SDL_Texture *ptr)
        {
            if (ptr)
                SDL_DestroyTexture(ptr);
        }
        void operator()(SDL_Renderer *ptr)
        {
            if (ptr)
                SDL_DestroyRenderer(ptr);
        }
        void operator()(SDL_Window *ptr)
        {
            if (ptr)
                SDL_DestroyWindow(ptr);
        }
        void operator()(SDL_RWops *ptr)
        {
            if (ptr)
                SDL_RWclose(ptr);
        }
    };

    // Shared pointer templated with deleter
    template <class T, class D = std::default_delete<T>>
    struct shared_ptr_with_deleter : public std::shared_ptr<T>
    {
        explicit shared_ptr_with_deleter(T *t = nullptr)
            : std::shared_ptr<T>(t, D()) {}

        void reset(T *t = nullptr)
        {
            std::shared_ptr<T>::reset(t, D());
        }
    };

    namespace Renderer
    {
        using UP = std::unique_ptr<SDL_Renderer, Deleter>;
        using SP = shared_ptr_with_deleter<SDL_Renderer, Deleter>;
    }

    namespace Texture
    {
        using UP = std::unique_ptr<SDL_Texture, Deleter>;
        using SP = shared_ptr_with_deleter<SDL_Texture, Deleter>;

        inline UP load(Renderer::SP const renderer, std::filesystem::path filepath)
        {
            return UP{IMG_LoadTexture(renderer.get(), filepath.c_str())};
        }

    }

    namespace Surface
    {
        using UP = std::unique_ptr<SDL_Surface, Deleter>;
        using SP = shared_ptr_with_deleter<SDL_Surface, Deleter>;
    }

    namespace Window
    {
        using UP = std::unique_ptr<SDL_Window, Deleter>;
        using SP = shared_ptr_with_deleter<SDL_Window, Deleter>;
    }

    namespace RWops
    {
        using UP = std::unique_ptr<SDL_RWops, Deleter>;
        using SP = shared_ptr_with_deleter<SDL_RWops, Deleter>;
    }
}
