#pragma once

#include <cmath>
#include <algorithm>

class Vector2D
{
public:
    /**
     * @brief The coordinates of the vector.
     */
    float x, y;

    Vector2D() : x(0.0f), y(0.0f) {}
    Vector2D(float x, float y) : x(x), y(y) {}

    /**
     * @brief Overloads the addition operator (+) to perform vector addition.
     *
     * @param other The vector to be added.
     * @return Vector2D A new Vector2D object representing the result of the addition.
     */
    inline Vector2D operator+(const Vector2D &other) const
    {
        return Vector2D(x + other.x, y + other.y);
    }

    /**
     * @brief Overloads the addition assignment operator (+=) to perform vector addition.
     *
     * @param other The vector to be added.
     * @return Vector2D& A reference to this Vector2D object after addition.
     */
    inline Vector2D &operator+=(const Vector2D &other)
    {
        x += other.x;
        y += other.y;
        return *this;
    }

    /**
     * @brief Overloads the subtraction operator (-) to perform vector subtraction.
     *
     * @param other The vector to be subtracted.
     * @return Vector2D A new Vector2D object representing the result of the subtraction.
     */
    inline Vector2D operator-(const Vector2D &other) const
    {
        return Vector2D(x - other.x, y - other.y);
    }

    /**
     * @brief Overloads the subtraction assignment operator (-=) to perform vector subtraction.
     *
     * @param other The vector to be subtracted.
     * @return Vector2D& A reference to this Vector2D object after subtraction.
     */
    inline Vector2D &operator-=(const Vector2D &other)
    {
        x -= other.x;
        y -= other.y;
        return *this;
    }

    /**
     * @brief Overloads the multiplication operator (*) to perform vector multiplication (coordinate by coordinate).
     *
     * @param other The vector to be multiplied with.
     * @return Vector2D A new Vector2D object representing the result of the coordinate-wise multiplication.
     */
    inline Vector2D operator*(const Vector2D &other) const
    {
        return Vector2D(x * other.x, y * other.y);
    }

    /**
     * @brief Overloads the multiplication assignment operator (*=) to perform vector multiplication (coordinate by coordinate).
     *
     * @param other The vector to be multiplied with.
     * @return Vector2D& A reference to this Vector2D object after coordinate-wise multiplication.
     */
    inline Vector2D &operator*=(const Vector2D &other)
    {
        x *= other.x;
        y *= other.y;
        return *this;
    }

    /**
     * @brief Overloads the division operator (/) to perform vector division (coordinate by coordinate).
     *
     * @param other The vector to be divided by.
     * @return Vector2D A new Vector2D object representing the result of the coordinate-wise division.
     */
    inline Vector2D operator/(const Vector2D &other) const
    {
        return Vector2D(x / other.x, y / other.y);
    }

    /**
     * @brief Overloads the division assignment operator (/=) to perform vector division (coordinate by coordinate).
     *
     * @param other The vector to be divided by.
     * @return Vector2D& A reference to this Vector2D object after coordinate-wise division.
     */
    inline Vector2D &operator/=(const Vector2D &other)
    {
        x /= other.x;
        y /= other.y;
        return *this;
    }

    /**
     * @brief Overloads the multiplication operator (*) to perform scalar multiplication.
     *
     * @param scalar The scalar value to multiply with.
     * @return Vector2D A new Vector2D object representing the result of the scalar multiplication.
     */
    inline Vector2D operator*(float scalar) const
    {
        return Vector2D(x * scalar, y * scalar);
    }

    /**
     * @brief Overloads the multiplication assignment operator (*=) to perform scalar multiplication.
     *
     * @param scalar The scalar value to multiply with.
     * @return Vector2D& A reference to this Vector2D object after scalar multiplication.
     */
    inline Vector2D &operator*=(float scalar)
    {
        x *= scalar;
        y *= scalar;
        return *this;
    }

    /**
     * @brief Overloads the division operator (/) to perform scalar division.
     *
     * @param scalar The scalar value to divide by.
     * @return Vector2D A new Vector2D object representing the result of the scalar division.
     */
    inline Vector2D operator/(float scalar) const
    {
        return Vector2D(x / scalar, y / scalar);
    }

    /**
     * @brief Overloads the division assignment operator (/=) to perform scalar division.
     *
     * @param scalar The scalar value to divide by.
     * @return Vector2D& A reference to this Vector2D object after scalar division.
     */
    inline Vector2D &operator/=(float scalar)
    {
        x /= scalar;
        y /= scalar;
        return *this;
    }

    inline bool operator<(Vector2D other)
    {
        return x < other.x && y < other.y;
    }

    /**
     * @brief Return the magnitude of the vector
     */
    inline float length() const
    {
        return std::sqrt(x * x + y * y);
    }

    /**
     * @brief Return the squared magnitude of the vector
     */
    inline float squared_length() const
    {
        return x * x + y * y;
    }

    /**
     * @brief Returns the result of scaling the vector to unit length.
     */
    inline Vector2D normalized() const
    {
        float l = length();
        if (l == 0.0f)
            return {0, 0};
        return (*this) / l;
    }

    /**
     * @brief Calculates the dot product of this vector and another vector.
     *
     * @param other The other vector for which the dot product is calculated.
     * @return float The dot product of this vector and the other vector.
     */
    inline float dot(const Vector2D &other) const
    {
        return x * other.x + y * other.y;
    }

    /**
     * @brief Calculates the angle of the vector relative to (1,0)
     *
     * @return The angle in radians
     */
    inline float angle() const
    {
        return std::atan2(y, x);
    }


    /**
     * @brief Returns the distance to another vector.
     */
    inline float distance_to(const Vector2D &other) const
    {
        return (other - *this).length();
    }

    /**
     * @brief Returns the squared distance to another vector.
     */
    inline float squared_distance_to(const Vector2D &other) const
    {
        return (other - *this).squared_length();
    }

    /**
     * @brief Returns the squared distance to the segment [ab].
     */
    inline float squared_distance_to_segment(Vector2D const &a, Vector2D const &b) const
    {
        Vector2D ab = (b - a);

        Vector2D projected = ab * std::clamp((*this - a).dot(ab) / ab.squared_length(), 0.0f, 1.0f);

        return (*this - a).squared_distance_to(projected);
    }

    /**
     * @brief Returns the distance to the segment [ab].
     */
    inline float distance_to_segment(Vector2D const &a, Vector2D const &b) const
    {
        return std::sqrt(squared_distance_to_segment(a, b));
    }

    /**
     * @brief Returns the perpendicular vector to this one.
     */
    inline Vector2D orthogonal() const
    {
        return {y, -x};
    }

    /**
     * @brief Returns the reflected vector on a line defined by @p normal
     *
     * @param normal The normal vector of the reflection line
     */
    inline Vector2D reflect(Vector2D const &normal) const
    {
        return normal * 2.0f * dot(normal) - *this;
    }

    friend std::ostream &operator<<(std::ostream &os, const Vector2D &vec)
    {
        os << "(" << vec.x << ", " << vec.y << ")";
        return os;
    }
};

/**
 * @brief Overloads the multiplication operator (*) to perform scalar multiplication when the scalar is on the left side.
 *
 * @param scalar The scalar value to multiply with.
 * @param v The vector to be multiplied.
 * @return Vector2D A new Vector2D object representing the result of the scalar multiplication.
 */
inline Vector2D operator*(float scalar, Vector2D const &v)
{
    return v * scalar;
}