#include "game.hpp"
#include "input.hpp"
#include "vector2d.hpp"

Game::Game(Window const &game_window) : game_window_{game_window},
                                        renderer_{game_window.get_renderer()},
                                        paddle_{renderer_},
                                        circle_guide_{SDL2::Texture::load(renderer_, "res/circle.svg")}
{
}

void Game::start()
{
    balls_.emplace_back(renderer_);
    balls_.front().speed = 250;
    balls_.front().set_position(game_window_.size() / 2.0f);
    circle_guide_.transform = Transform2D::translation(game_window_.size() / 2.0f);

    paddle_.pivot_point_ = game_window_.size() / 2.0f;
}

void Game::update()
{
    // Fixed time-step update
    // https://youtu.be/lW6ZtvQVzyg?si=q26fGTe4RbFgwULH

    static uint64_t time_accumulator_ms = 0;
    static uint64_t ticks_last_ms = 0;
    static uint64_t const slice_ms = 10;

    uint64_t delta = SDL_GetTicks64() - ticks_last_ms;
    ticks_last_ms += delta;
    time_accumulator_ms += delta;

    while (time_accumulator_ms > slice_ms)
    {

        float delta_s = slice_ms / 1000.0f;

        update(delta_s);

        for (auto &&ball : balls_)
        {
            ball.update(delta_s);
        }

        time_accumulator_ms -= slice_ms;
    }
}

void Game::update(float delta)
{
    static Vector2D direction{0, 1};

    paddle_.update(delta);

    auto &ball = balls_[0];
    Vector2D target_move = ball.speed * direction * delta;
    Vector2D collision_normal{};

    Vector2D movement = target_move;

    if (ball.collide_movement(paddle_, movement, collision_normal))
    {
        ball.set_position(ball.position() + movement);

        direction = direction.reflect(collision_normal.orthogonal());
        movement = direction * (target_move - movement).length();
        target_move = movement;
    }
    else
    {
        ball.set_position(ball.position() + movement);
    }
}

void Game::render()
{
    circle_guide_.draw(renderer_);

    for (auto &&ball : balls_)
    {
        ball.render();
    }

    paddle_.render();
}